/* 
baby box  27.11.2018

- ILI9341 touch display (2.4") - NOTE: requires 3.3v logic!  
- MFRC522 RFID module
- DFRobotDFPlayer module
- Arduino Nano/Due

Required Arduino libraries:
  1. Install Adafruit GFX library (Arduino IDE->Sketch->Add library->Library manager)
  2. Install Adafruit ILI9341  library  (Arduino IDE->Sketch->Add library->Library manager)


Wiring:

  Arduino Nano/Due        module
  D5    ----------------- powerOffSwitch (circuit)                    
  5V    ----------------- MPU6050 VCC
  GND   ----------------- MPU6050 GND
  A5    ----------------- MPU6050 SCL   
  A4    ----------------- MPU6050 SDA            
  5.0v  ----------------- ILI9341 VCC
  3.3v  ----R50 ohm------ ILI9341 LED
  GND   ----------------- ILI9341 GND      
  A0    --LVC------------ ILI9341 RESET  (LVC = 5v->3.3v level converter)
  D4    --LVC------------ ILI9341 DC
  D6    --LVC------------ ILI9341 CS
  D7    --LVC------------ ILI9341 T_CS
  D3    --LVC------------ ILI9341 T_IRQ      
  D13/SCK  --LVC-----|--- ILI9341 SCK        (for Arduino Due SCK, MISO, MOSI pins: https://www.arduino.cc/en/Reference/SPI )  
                     x--- ILI9341 T_CLK  
                     x--- MFRC522 SCK
  D12/MISO --LVC-----|--- ILI9341 MISO 
                     x--- ILI9341 T_DO (=MISO) 
                     x--- MFRC522 MISO
  D11/MOSI --LVC-----|--- ILI9341 MOSI
                     x--- ILI9341 T_DIN (=MOSI)
                     x--- MFRC522 MOSI
  D9    ----------------- MFRC522 RST
  D10   ----------------- MFRC522 SDA/SS
  GND   ----------------- MFRC522 GND
  3.3V  ----------------- MFRC522 3.3V
  5V    ----------------- DFRobotDFPlayer VCC 
  D8 /TX1 --------------- DFRobotDFPlayer RX  
  D2 /RX1 --------------- DFRobotDFPlayer TX  
  GND   ----------------- DFRobotDFPlayer GND 
                          DFRobotDFPlayer SPK_1 -- 25 Ohm speaker  
                          DFRobotDFPlayer SPK_2 -- 25 Ohm speaker
*/

#include <SPI.h>
#include "Adafruit_ILI9341.h"
#include "XPT2046_Touchscreen.h"
#include "MFRC522.h"
#include "MFRC522Hack.h"
#ifdef __SAM3X8E__
  #include "DueFlashStorage.h"
#else
  #include <EEPROM.h>
  #include <SoftwareSerial.h>
#endif
#include "DFRobotDFPlayerMini.h"
#include "MPU6050.h"

// ------- config ---------------------------------

// Arduino pins
#define pinPowerOffSwitch 5 // for power-off circuit (optional)
// RFID
#define RFID_SS_PIN 10
#define RFID_RST_PIN 9
// touch display
#define TFT_DC 4
#define TFT_CS 6
#define TFT_RST A0
#define TOUCH_CS 7
#define TOUCH_IRQ 3

#define DFPLAYER_TX 2 // for Nano only
#define DFPLAYER_RX 8 // for Nano only
#define DFPLAYER_Serial Serial1 // for Due only


// screen size (pixels)
#define TFT_W 320
#define TFT_H 240

// touch min/max values
#define TS_MINX 470
#define TS_MINY 250
#define TS_MAXX 3900
#define TS_MAXY 3900

//Set default volume value. From 0 to 30
int volume = 15;

// maximum song and tag count
#define SONG_COUNT  50
#define TAG_COUNT  20

#define VERSION "27.11.2018"

// ----------------------------------------

// tag status
#define TAG_NOT_PRESENT  0
#define TAG_MEM_FULL     1
#define TAG_NEW          2
#define TAG_PROGRAMMED   3


// list of all TAG IDs 
uint16_t tagList[TAG_COUNT] = {0};
// list of corresponding song numbers for each TAG in tagList
uint8_t assignedSong[TAG_COUNT] = {0};

//MPU6050 mpu;

MFRC522 mfrc522(RFID_SS_PIN, RFID_RST_PIN);  // Create MFRC522 instance.
MFRC522Hack mfrc522Hack(&mfrc522);  // Create MFRC522Hack instance.

XPT2046_Touchscreen ts(TOUCH_CS, TOUCH_IRQ);
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);


DFRobotDFPlayerMini myDFPlayer;

#ifdef __SAM3X8E__
  // Due flash
  DueFlashStorage EEPROM;
  #define EEPROM_STEP_SIZE 4 // bytes to next EEPROM address (must be aligned by 4 for Due)
#else  
  SoftwareSerial mySoftwareSerial(DFPLAYER_TX, DFPLAYER_RX); // RX, TX
  #define EEPROM_STEP_SIZE 1 // bytes to next EEPROM address
#endif

int count = 0;
unsigned long activityTimeout = 0;
unsigned long saveTimeout = 0;
int playIdx = 0;
uint16_t lastTag = 0;
uint16_t currTag = 0;
unsigned long stopTimeout = 0;
unsigned long touchTimeout = 0;
unsigned long nextTouchTime = 0;
unsigned long powerOffTimeout = millis() + 20000; // initial power-off timeout (20 secs)
unsigned long firstTouchTime = 0;
unsigned long nextCardTime = 0;
bool locked = false;
bool playing = false;
Stream *playerSerial;
byte tagState = TAG_NOT_PRESENT;


void triggerSave(){
  saveTimeout = millis() + 5000;  
}

// print song for each TAG ID 
void printTags(){  
  Serial.println(F("-----------------"));
  for (int i=0; i < TAG_COUNT; i++){    
    Serial.print(F("  TAG "));
    Serial.print(tagList[i]);
    Serial.print(F(" => "));
    Serial.print(F("song "));
    Serial.println(assignedSong[i]+1);        
  }    
  Serial.println(F("-----------------"));
}

// clear TAG ID table
void clearEEPROM(){
  Serial.print(F("clearing EEPROM..."));        
  unsigned long nextTime = 0;
  for (int addr=0; addr < 256; addr++) {    
    EEPROM.write(addr, 0);      
  }
  for (int i=0; i < TAG_COUNT; i++) {
    tagList[i]=0;
    assignedSong[i]=0;
  }
  Serial.println(F("done"));
  printTags();
}

// save TAG ID table
void saveEEPROM(){
  Serial.print(F("saveEEPROM..."));
  float floatValue;
  int addr = 0;  
  byte magic = 0xAA;
  EEPROM.write(addr, magic);
  addr += EEPROM_STEP_SIZE;
  EEPROM.write(addr, volume);
  addr += EEPROM_STEP_SIZE;
  EEPROM.write(addr, locked);
  addr += EEPROM_STEP_SIZE;
  for (int i=0; i < TAG_COUNT; i++){
    EEPROM.write(addr, tagList[i] >> 8); 
    addr += EEPROM_STEP_SIZE;
    EEPROM.write(addr, tagList[i] & 0xFF); 
    addr += EEPROM_STEP_SIZE;
  }  
  for (int i=0; i < TAG_COUNT; i++){
    EEPROM.write(addr, assignedSong[i]); 
    addr += EEPROM_STEP_SIZE;
  }    
  Serial.println(F("done"));
}

// load TAG ID table
void loadEEPROM(){
  Serial.print(F("loadEEPROM..."));
  for (int i=0; i < TAG_COUNT; i++) {
    tagList[i]=0;
    assignedSong[i]=0;
  }
  int addr = 0;
  unsigned short magic = 0;  
  magic = EEPROM.read(addr);
  if (magic != 0xAA) {
    tft.println(" NO DATA!");
    Serial.println(F("no valid EEPROM data found"));
    return;
  }
  Serial.println(F("valid EEPROM data found"));
  tft.println(" OK");
  addr += EEPROM_STEP_SIZE;
  volume = EEPROM.read(addr);
  addr += EEPROM_STEP_SIZE;
  locked = EEPROM.read(addr);
  addr += EEPROM_STEP_SIZE;
  for (int i=0; i < TAG_COUNT; i++){
    tagList[i] = (EEPROM.read(addr) << 8) | EEPROM.read(addr+EEPROM_STEP_SIZE);    
    addr += EEPROM_STEP_SIZE*2;
  }  
  for (int i=0; i < TAG_COUNT; i++){
    assignedSong[i] = (EEPROM.read(addr));
    addr += EEPROM_STEP_SIZE;
  }       
  Serial.println(F("done"));
}

void displayText(){  
  tft.setCursor(0,20);  
  tft.setTextSize(2);
  tft.print(F("Vol"));    

  tft.setCursor(0,80);
  tft.setTextSize(2);
  tft.print(F("Song"));     

  tft.setCursor(0,140);
  tft.setTextSize(2);
  tft.print(F("Tag"));   
}

void displayLocked(){  
  tft.setCursor(60,200);  
  tft.setTextSize(2);
  tft.print(F("Press long to unlock"));    
  tft.setTextSize(5);
  tft.drawRect(0, 180, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(10,190);
  tft.print("#");  
}

void displaySetup(){  
  // volume buttons -, +
  tft.setTextSize(5);
  tft.drawRect(120, 0, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(130,10);
  tft.print("-");
  tft.drawRect(180, 0, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(190,10);
  tft.print("+");

  // song buttons -, +, pause   
  tft.drawRect(120, 60, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(130,70);
  tft.setTextSize(5);   
  tft.print("-");
  tft.drawRect(180, 60, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(190,70);
  tft.print("+");  
  tft.drawRect(240, 60, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(250,75);
  tft.print("\"");

  // lock button
  tft.drawRect(0, 180, 50, 50, ILI9341_YELLOW);  
  tft.setCursor(10,190);
  tft.print("#");

  tft.setCursor(60,200);  
  tft.setTextSize(2);
  tft.print(F("Press long to lock"));    
}

void redrawDisplay(){
  tft.fillScreen(ILI9341_BLACK);        
  displayText();
  if (locked) displayLocked();
    else displaySetup();  
  updateDisplayData();
}

void updateDisplayData(){
  tft.fillRect(60, 20, 40, 20, ILI9341_BLACK);
  tft.setCursor(60,20);
  tft.setTextSize(2);
  tft.print(volume);   

  tft.fillRect(60, 80, 40, 20, ILI9341_BLACK);
  tft.setCursor(60,80);
  tft.setTextSize(2);
  tft.print(playIdx+1);     

  tft.fillRect(60, 140, 250, 20, ILI9341_BLACK);
  tft.setCursor(60,140);
  tft.setTextSize(2);
  tft.print(currTag);       
  tft.print(" (");       
  switch (tagState){
    case TAG_NOT_PRESENT: tft.print(F("not present")); break;
    case TAG_MEM_FULL: tft.print(F("error-mem full")); break;
    case TAG_PROGRAMMED: tft.print(F("programmed")); break;
    case TAG_NEW: tft.print(F("new")); break;
  }  
  tft.print(")");           
}


void setup() {  
  pinMode(pinPowerOffSwitch, OUTPUT);
  digitalWrite(pinPowerOffSwitch, HIGH);
  
  Serial.begin(115200); // Initialize serial communications with the PC  
  Serial.print(F("babybox "));
  Serial.println(F(VERSION));
  
  SPI.begin();      // Init SPI bus

  tft.begin(); 
  tft.setRotation(1);
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextColor(ILI9341_YELLOW);      
  tft.setTextSize(2);
  tft.setCursor(0, 0);
  tft.print(F("babybox "));
  tft.println(VERSION);
    
  tft.print(F("loading EEPROM-"));
  loadEEPROM();
  
  tft.println(F("starting touch"));
  
  ts.begin();
  ts.setRotation(1);  
  
  #ifdef __SAM3X8E__
    playerSerial = &DFPLAYER_Serial;    
    DFPLAYER_Serial.begin(9600);
  #else
    playerSerial = &mySoftwareSerial;        
    mySoftwareSerial.begin(9600);
  #endif     
  
  //randomSeed(analogRead(1));    

  Serial.println(F("mfrc522 init..."));
  
  tft.println(F("starting RFID"));
  
  mfrc522.PCD_Init(); // Init MFRC522 card
  /*while (!mfrc522.PCD_PerformSelfTest()){
    Serial.println(F("Could not find MFRC522, check wiring!"));
    delay(500);
  }*/  
  
  tft.println(F("starting DFPlayer"));
  
  while (!myDFPlayer.begin(*playerSerial)) {  //Use softwareSerial to communicate with mp3.
    Serial.println(F("Unable to begin DFRobotDFPlayerMini:"));
    Serial.println(F("1.Please recheck the connection!"));
    Serial.println(F("2.Please insert the SD card!"));
    delay(5000);
  }
  Serial.println(F("DFPlayer Mini online."));  
  

  myDFPlayer.setTimeOut(500); //Set serial communictaion time out 500ms
  //myDFPlayer.volume(VOLUME);  //Set volume value. From 0 to 30
  myDFPlayer.EQ(DFPLAYER_EQ_NORMAL);
  myDFPlayer.outputDevice(DFPLAYER_DEVICE_SD);
  
  while (count == 0){
    Serial.print(F("files:"));    
    count = SONG_COUNT; //myDFPlayer.readFileCounts();
    Serial.println(count);  
    delay(1000);    
  }
  
  printTags();
       
  //myDFPlayer.play(playIdx);  //Play the first mp3  
 
  Serial.println(F("Scan PICC to see UID and type..."));
    
  setVolume(volume);      
  redrawDisplay();
}


// pause playing song
void pause(){
  Serial.println(F("PAUSE"));
  myDFPlayer.pause();  
  playing = false;  
  updateDisplayData();
}

// unpause pause command
void unpause(){
  Serial.println(F("UNPAUSE"));
  myDFPlayer.start();  
  playing = true;    
  updateDisplayData();
}

void setLocked(bool flag){
  Serial.print(F("LOCKED "));
  Serial.println(flag);
  locked = flag;
  triggerSave();
}

void setVolume(int vol){
  volume = vol;
  Serial.print(F("VOLUME "));
  Serial.println(vol);  
  myDFPlayer.volume(vol);  //Set volume value. From 0 to 30
  updateDisplayData();  
}

// play song with idx
void play(int idx){
  playIdx = idx;  
  Serial.print(F("PLAYING "));
  Serial.println(idx+1);  
  myDFPlayer.play(idx+1);        
  playing = true;  
  powerOffTimeout = millis() + 600000; // reset power-off timeout (600 secs)
  updateDisplayData();
}

// play song with next idx
void playNext(){
  playIdx++;
  if (playIdx>= SONG_COUNT) playIdx = 0;
  play(playIdx);
}

// play song with previous idx
void playPrevious(){
  playIdx--;
  if (playIdx < 0) playIdx= SONG_COUNT-1;      
  play(playIdx);
}

// assign TAG to song number
void assignSong(uint16_t tag, byte playIdx){
  int idx = -1;      
  for (int i=0; i < TAG_COUNT; i++){
    if (tagList[i] == 0) idx = i; // remember free index
    if (tagList[i] == tag) { // found tag in list     
      idx = i;
      break; 
    }
  }  
  // TAG not yet in list => add it to free index  
  if (idx == -1){
    printTags();
    Serial.println(F("WARNING: cannot assign song - no free index!"));
    tagState = TAG_MEM_FULL;
    return;       
  }
  tagList[idx]=tag;
  assignedSong[idx] = playIdx;  
  tagState = TAG_PROGRAMMED;
  printTags();  
  triggerSave();
}

// find assigned song numbre for TAG
int findAssignedSong(uint16_t tag){
  for (int i=0; i < TAG_COUNT; i++){
    if (tagList[i] == tag) {  // found tag in list            
      Serial.println(F("TAG is programmed"));      
      tagState = TAG_PROGRAMMED;           
      return assignedSong[i];
    }
  }
  Serial.println(F("TAG not programmed"));        
  tagState = TAG_NEW;  
  return 0;  
}

void loop() {               
  // --------- serial console ------------------------------
  if (Serial.available()){
    // serial console test functions
    char ch = Serial.read();
    if (ch == 'c') {
      clearEEPROM();
    }
    if (ch == '+') {
      playNext();
      stopTimeout = millis() + 1000000;
    }
    if (ch == '-') {
      playPrevious();      
      stopTimeout = millis() + 1000000;      
    }
    if (ch == 'r') {
      // Set new UID to card            
      byte newUid[] = {0x00, 0x00, 0x00, 0x00};
      newUid[0] = random(0, 256);                
      newUid[1] = random(0, 256);                
      newUid[2] = random(0, 256);                
      newUid[3] = random(0, 256);                      
      if ( mfrc522Hack.MIFARE_SetUid(newUid, (byte)4, true) ) {
        Serial.println(F("Wrote new UID to card."));
        mfrc522.PICC_HaltA();
      }
    }   
  }

  if (millis() > stopTimeout){
    if (playing) {      
      pause();       // pause playing
    }    
  }

  // ----- RFID reader --------------------------------------      
  // Look for new cards
  if (millis() > nextCardTime){
    nextCardTime = millis() + 400;
    if ( mfrc522.PICC_IsNewCardPresent()) {  
      // Select one of the cards
      if ( mfrc522.PICC_ReadCardSerial()) {
        // make unique key based on serial number 
        currTag = 0;
        for (byte i = 0; i < mfrc522.uid.size; i++) {
          currTag += (i+1) * mfrc522.uid.uidByte[i];    
        }       
        // check if there is a new unique key            
        if (currTag != lastTag){        
          printTags();
          int idx = findAssignedSong(currTag);              
          lastTag = currTag;
          Serial.print(F("new TAG: "));
          Serial.println(currTag);                        
          //printTags();        
          //playIdx = random(1, count+1);                
          play(idx); // play new song                
          if (locked) redrawDisplay(); // clear display on new card
        } else { // same card again
          if (millis() > stopTimeout) { // unpause playing      
            unpause();          
          }
        }  
        stopTimeout = millis() + 1000;    
      }      
    } else {
      // no card present
      if (millis() > stopTimeout) {
        // TAG not present
        currTag = 0;      
        tagState = TAG_NOT_PRESENT;
      }    
    }
  }

  
  // ------ touch display --------------------  
  if (millis() > nextTouchTime){    
    nextTouchTime = millis() + 50;
    boolean istouched = false;
    if (ts.tirqTouched()) {
      istouched = ts.touched();
    }  
    
    if (istouched)  {    
      TS_Point p = ts.getPoint();
      // map from 0->1023 to tft.width
      int px = map(p.x, TS_MINX, TS_MAXX, TFT_W, 0);
      int py = TFT_H - map(p.y, TS_MINY, TS_MAXY, TFT_H, 0);          
      
      if (locked){
        // locked: no button actions, allow drawing
        tft.fillCircle(px, py, 3, ILI9341_GREEN);            
      } 
      
      if (millis() > touchTimeout)  {    
        touchTimeout = millis() + 300;
        
        if (!locked){        
          // unlocked:  button actions
          if ((px >= 120) && (py >= 60) && (px <= 120+50) && (py <= 60+50)){
            // play previous          
            playPrevious();
            if (currTag != 0) assignSong(currTag, playIdx);  
            updateDisplayData();
            stopTimeout = millis() + 1000000;       
            touchTimeout = millis() + 1000;
          }      
          if ((px >= 180) && (py >= 60) && (px <=180+50) && (py <= 60+50)) {
            // play next          
            playNext();
            if (currTag != 0) assignSong(currTag, playIdx);  
            updateDisplayData();
            stopTimeout = millis() + 1000000;            
            touchTimeout = millis() + 1000;
          }      
          if ((px >= 240) && (py >= 60) && (px <=240+50) && (py <= 60+50)) {
            // pause/play        
            if (playing) pause();
              else unpause();          
          }     
          if ((px >= 120) && (py >= 0) && (px <= 120+50) && (py <= 50)){
            // decrease volume
            setVolume(max(0, volume-1));                
            triggerSave();
          }
          if ((px >= 180) && (py >= 0) && (px <= 180+50) && (py <= 50)){
            // increase volume
            setVolume(min(30, volume+1));        
            triggerSave();
          }
        }
        
        if ((px >= 0) && (py >= 180) && (px <= 50) && (py <= 180+50)){
          // lock/unlock
          if (millis() > firstTouchTime + 3000){
            setLocked( !locked );                           
            firstTouchTime = millis();    
            redrawDisplay();                            
          }               
        }
          
      }
    } else {
      touchTimeout = 0;
      firstTouchTime = millis();
    }
  }
      

  // ------- power-off circuit ---------------
  if (millis() > powerOffTimeout){
    // no new song after timeout => switch off player
    powerOffTimeout = millis() + 20000;
    Serial.println(F("powering off..."));        
    digitalWrite(pinPowerOffSwitch, LOW);
  }

  // ------- saving ---------------------------
  if ((saveTimeout != 0) && (millis() > saveTimeout))  {
    saveTimeout = 0;
    saveEEPROM();    
  }
  
}



